// Generated from Flocc program
// see http://www.flocc.net/
// (was called DistHist2a.cpp)

#include <mpi.h>
#include <vector>
#include <cstdlib>
#include <memory>
#include <time.h> // to see srand
#include <sparsehash/dense_hash_map> // hashmap
#include <cstring> // for strcmp
#include <cstdio> // sprintf
#include <climits> // for hashmap empty key values
#include <cmath> // for hashmap empty key values

// RAND NUMBERS
#define randFloat() ((double)rand()/RAND_MAX)

// BASE TYPES 

typedef struct nullStruct {
  bool operator<(const nullStruct& b) const {
    return false;
  }
  bool operator==(const nullStruct& b) const {
    return true;
  }
} nullStruct;

// PRINTING

void printVal(int v) { printf("%i", v); }
void printVal(double v) { printf("%f", v); }
void printVal(const char *v) { printf("%s", v); }
void printVal(nullStruct v) {
  printVal("null");
}

// HASHING

using google::dense_hash_map;

unsigned hashData ( const void *key, int len )
{
    unsigned char *p = (unsigned char*)key;
    unsigned h = 0;
    int i;
  
    for ( i = 0; i < len; i++ ) {
      h += p[i];
      h += ( h << 10 );
     h ^= ( h >> 6 );
   }
 
   h += ( h << 3 );
   h ^= ( h >> 11 );
   h += ( h << 15 );
 
   return h;
}

template <class T>
struct hash {
  public:
  inline size_t operator()(const T& x) const {
    std::cerr << "Called unimplemented hash<T> class!\n";
    throw -1;
  }
};

template <>
struct hash<int> {
  public:
  inline size_t operator()(const int& x) const {
    return x;
  }
};
const hash<int> intHasher;

template <>
struct hash<bool> {
  public:
  inline size_t operator()(const bool& x) const {
    return x;
  }
};
const hash<bool> boolHasher;

// empty keys for hashmaps
// TODO generate empty values for all structs...
const int intEmpty = INT_MIN;
const double doubleEmpty = NAN; 

// EQUALITY

struct eqcharStar
{
  bool operator()(const char* s1, const char* s2) const
  {
    return (s1 == s2) || (s1 && s2 && strcmp(s1, s2) == 0);
  }
};

struct eqint
{
  bool operator()(const int s1, const int s2) const
  {
    return (s1 == s2);
  }
};

struct eqdouble
{
  bool operator()(const double s1, const double s2) const
  {
    // TODO - change so equal to within an interval.
    return (s1 == s2);
  }
};

struct eqbool
{
  bool operator()(const bool s1, const bool s2) const
  {
    return (s1 == s2);
  }
};

// MPI STUFF

typedef struct NodeLayout {
  // all partition root nodes
  // in this node set
  MPI::Comm *partitionCom;
  int partitionCount, // number of partitions
      partitionRank;  // partition rank of this node

  // all nodes to mirror on in this partition
  MPI::Comm *mirrorCom;
  int mirrorCount, // number of nodes in mirror set
      mirrorRank,  // rank in mirror set of this node
      mirrorRootRank;    // rank of root in mirror set

  // if mirrorRank == mirrorRootRank, then this is a partition
  // head/root node

  // if partitionRank or mirrorRank is -1
  // then this node is not in this node layout

  bool operator==(const NodeLayout& b) const {
    // comparing layouts on a single machine, it should
    // be sufficient to compare communicators, as no other
    // layout should exist ON THIS MACHINE with the same
    // communicators.
    return (partitionCom == b.partitionCom) &&
           (mirrorCom == b.mirrorCom);
  }

  void Free() {
    // free communicators, 
    // ONLY SAFE AT VERY END, as other
    // layouts might use one of the same communicators.
    if (partitionCom != NULL && (*partitionCom) != MPI::COMM_WORLD) partitionCom->Free();
    if (mirrorCom != NULL) mirrorCom->Free();
  }

} NodeLayout;

NodeLayout allNodes;

#define NL_ON_FRINGE(nl) ((nl.mirrorRank == nl.mirrorRootRank) && (nl.mirrorRank >= 0))

// converts the rank of a process in one communicator to it's rank
// in another communicator.
void translateRank(MPI::Comm *srcComm, int srcRank, MPI::Comm *destComm, int *destRank) {
  MPI_Group orig_group, new_group; 
  orig_group = srcComm->Get_group();
  new_group = destComm->Get_group();
  MPI::Group::Translate_ranks(orig_group, 1, &srcRank, new_group, destRank);
} 

//typedef void MPI::Comm::Errhandler_fn(MPI::Comm &, int *, ... );
void throwErrHandler(MPI::Comm &comm, int *errCode, ...) {
  int len = 5000;
  char str[len];
  for (int i = 0; i < len; i++) str[i] = 0;
  MPI::Get_error_string(*errCode, str, len);
  std::cerr << "MPI Error: " << str << "\n";
  throw 1;
}

// allToAll communicate vectors
template <class T>
void exchangePartitions(
  std::vector<T> *inputs, 
  std::vector<T> *output, 
  NodeLayout *layout) {
  // if just one node, return
  const int pcount = layout->partitionCount;
  if (pcount <= 1) { 
    *output = inputs[0];
    return;
  } 

  std::cout << "MAX VECTOR SIZE: " << output->max_size() << "\n\n";

  // alltoall data 
  // inputs 228
  // outV 225 = output
  int sendCnts[pcount]; // 229
  int recvCnts[pcount]; // 230
  int sendDispls[pcount]; // 231
  int recvDispls[pcount]; // 232 

  // send/recv partition sizes
  for (int pid = 0; pid < pcount; pid++) {
    sendCnts[pid] = inputs[pid].size() * sizeof(T);  // sendCnts
    void* ptr = &inputs[pid].front();
    std::cout << "Pointer: " << ptr << "\n\n";
    sendDispls[pid] = MPI::Get_address(ptr); // sendDispls
  }
  //return;
  layout->partitionCom->Alltoall(sendCnts, 1, MPI_INT, recvCnts, 1, MPI_INT);

  // sum counts, and make displacements
  int count; // 233      
  for (int pid = 0; pid < pcount; pid++) {
    recvDispls[pid] = count; // recv displs
    count += recvCnts[pid]; // total bytes to recv
  }
  count /= sizeof(T); // total elements to recv
  output->resize(count); // reserve room for total elements

  // send/recv partitions
  layout->partitionCom->Alltoallv(MPI_BOTTOM, sendCnts, sendDispls, MPI_PACKED,
        &(output->front()), recvCnts, recvDispls, MPI_PACKED);
}

// TIME

#define MAX_TIME_HEADER 1024
int timeRank;
char timeHeader[MAX_TIME_HEADER];
double beginTime, middleTime;

void initTime() {
  // rank
  timeRank = MPI::COMM_WORLD.Get_rank();

  // start time
  beginTime = MPI_Wtime();
  middleTime = beginTime;
}

void logTime(const char* event, double *startTime) {
  if (timeRank == 0) {
    double End = MPI_Wtime();
    std::cerr << timeHeader << "\t" 
              << event << "\t" << (End - (*startTime)) << std::endl;
  }
  middleTime = MPI_Wtime();
}

//int main(int argc, char** argv) {}
// code generation dist histogram



// declaring node layouts
NodeLayout nl1;


typedef struct struct220 {
int v0;
double v1;
bool operator<(const struct220& b) const {
  return (v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1)));
}
bool operator>=(const struct220& b) const {
  return !((v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1))));
}
inline bool operator==(const struct220& b) const {
  return (v0 == b.v0) && (v1 == b.v1);
}
inline bool operator!=(const struct220& b) const {
  return !((v0 == b.v0) && (v1 == b.v1));
}
} struct220;
void printVal(const struct220& v) {
printVal("(");
printVal(v.v0);
printVal(", ");
printVal(v.v1);
printVal(")");

}
// HASH NEEDS C++ 11 SO OMITTED FOR NOW
struct eqstruct220 {
inline bool operator()(const struct220& a, const struct220& b) const {
  return (a.v0 == b.v0) && (a.v1 == b.v1);
}
};


 // user def reduction function
  void userFun16(const void* invec, void *inoutvec, int len, const MPI::Datatype& datatype) {
   const struct220* in = (const struct220*)invec;
   struct220* out = (struct220*)inoutvec;
   // TODO: WHEN SENDING PROPER DATATYPE, CHANGE (len / sizeof(struct220)) to len!
   for (int i = 0; i < (len / sizeof(struct220)); i++) {
     struct220 val2 = out[i];
         double v218;
    v218 = (in[i].v1 < val2.v1) ? in[i].v1 : val2.v1;
    double v219;
    v219 = (in[i].v0 < val2.v0) ? val2.v0 : in[i].v0;
  // TupNd assignment
    out[i].v0 = v219;
    out[i].v1 = v218;
    
  
  }
 }

MPI::Op v221;

typedef struct struct239 {
int v0;
int v1;
bool operator<(const struct239& b) const {
  return (v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1)));
}
bool operator>=(const struct239& b) const {
  return !((v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1))));
}
inline bool operator==(const struct239& b) const {
  return (v0 == b.v0) && (v1 == b.v1);
}
inline bool operator!=(const struct239& b) const {
  return !((v0 == b.v0) && (v1 == b.v1));
}
} struct239;
void printVal(const struct239& v) {
printVal("(");
printVal(v.v0);
printVal(", ");
printVal(v.v1);
printVal(")");

}
// HASH NEEDS C++ 11 SO OMITTED FOR NOW
struct eqstruct239 {
inline bool operator()(const struct239& a, const struct239& b) const {
  return (a.v0 == b.v0) && (a.v1 == b.v1);
}
};


typedef struct struct261 {
double v0;
double v1;
bool operator<(const struct261& b) const {
  return (v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1)));
}
bool operator>=(const struct261& b) const {
  return !((v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1))));
}
inline bool operator==(const struct261& b) const {
  return (v0 == b.v0) && (v1 == b.v1);
}
inline bool operator!=(const struct261& b) const {
  return !((v0 == b.v0) && (v1 == b.v1));
}
} struct261;
void printVal(const struct261& v) {
printVal("(");
printVal(v.v0);
printVal(", ");
printVal(v.v1);
printVal(")");

}
// HASH NEEDS C++ 11 SO OMITTED FOR NOW
struct eqstruct261 {
inline bool operator()(const struct261& a, const struct261& b) const {
  return (a.v0 == b.v0) && (a.v1 == b.v1);
}
};




void run() {v221.Init(&userFun16, true);

  // initializing node layouts
  nl1.partitionCom = &MPI::COMM_WORLD;
  nl1.partitionCount = MPI::COMM_WORLD.Get_size();
  nl1.partitionRank = MPI::COMM_WORLD.Get_rank();
  nl1.mirrorCom = NULL;
  nl1.mirrorCount = 1;
  nl1.mirrorRank = 0;
  nl1.mirrorRootRank = 0;
allNodes = nl1;


  int v2;
  v2 = 1073741824;
  int v223;
  v223 = 100;
  int v224;
  v224 = 1;
  logTime("init", &middleTime);
// genDistRandFloats: 
  std::vector<double > v7;
    double v5;
    int v6;
  int v3 = v2 % nl1.partitionCount;
int v4 = v2 / nl1.partitionCount;
if (nl1.partitionRank < v3) {
  v3 = (nl1.partitionRank * v4) + nl1.partitionRank;
  v4++;
} else {
  v3 += nl1.partitionRank * v4;
}
v7.reserve(v4);
for (v6 = v3; v6 < (v3 + v4); v6++) {
  v5 = randFloat();
  v7.push_back(v5);
}
logTime("genfloats", &middleTime);

// readDistArr1D
// readDistArr1D:init
  // reduceDistStream:initStream
    struct261 v14;
    
 if (NL_ON_FRINGE(nl1)) {
  // calc start index for this node
    int v9;
   // declare readDistArr1D index
  // TODO THIS IS BROKEN, NEEDS TO PASS OFFSET AROUND WITH VECTOR 
  // BECAUSE SOMETIMES VECTOR WON'T DIVIDE EVENLY
  v9 = nl1.partitionRank * v7.size();
    std::vector<double >::iterator v8;
   // declare readDistArr1D iterator
  for (v8 = v7.begin(); v8 != v7.end(); ++v8) {
  // begin readDistVec consume
    // reduceDistStream:consumers 
  // decVal1
    double v10;
    double v11;
    
  // decVal2
    double v12;
    double v13;
    
  // projVal
    // TupNd assignment
      v10 = (*v8);
      v11 = (*v8);
      
    
  // combineVals
      double v281;
      v281 = (v14.v1 < v11) ? v14.v1 : v11;
      double v282;
      v282 = (v14.v0 < v10) ? v10 : v14.v0;
    // TupNd assignment
      v12 = v282;
      v13 = v281;
      
    
  // copyRes
    v14.v0 = v12;
    v14.v1 = v13;
    
  
  // end readDistVec consume
  v9++; // next index
  }
}
  // reduceDistStream:finStream
  // allreduce among all nodes
    struct261 v15;
      v15 = v14;
    
  nl1.partitionCom->Allreduce(&v15, &v14, sizeof(struct220), MPI_PACKED, v221);
   // finish readDistArr1D

  double v222;
  v222 = v14.v0 - v14.v1;
  double v225;
  v225 = v223 - v224;
  double v226;
  v226 = (double)v225;
  double v227;
  v227 = v226 / v222;
// readDistArr1D
// readDistArr1D:init
    // declare hashmap
       dense_hash_map<int, int, hash<int >, eqint > v236;
       // TODO init size
    v236.set_empty_key(intEmpty);
    
 if (NL_ON_FRINGE(nl1)) {
  // calc start index for this node
    int v229;
   // declare readDistArr1D index
  // TODO THIS IS BROKEN, NEEDS TO PASS OFFSET AROUND WITH VECTOR 
  // BECAUSE SOMETIMES VECTOR WON'T DIVIDE EVENLY
  v229 = nl1.partitionRank * v7.size();
    std::vector<double >::iterator v228;
   // declare readDistArr1D iterator
  for (v228 = v7.begin(); v228 != v7.end(); ++v228) {
  // begin readDistVec consume
    // mapDistStream:
  //decOuts
    int v230;
    int v231;
    //kfCode
    // identity fun assignment
      v230 = v229;
      
    //vfCode
      double v293;
      v293 = (*v228) * v227;
    v231 = (int)v293;
    
  //consume:
    //begin groupReduceDistStreamEx consumer
      int v232;
      
      int v233;
      
      int v234;
      
      int v235;
      
      // TupAccNd assignment
        v232 = v231;
        
      
      v233 = 1;
      
    if (v236.count(v232)) {
      v234 = v236[v232];
        v235 = v233 + v234;
      
      v236[v232] = v235;
    } else {
      v236[v232] = v233;
    }
    //end groupReduceDistStreamEx consumer
    
  
  // end readDistVec consume
  v229++; // next index
  }
}
  logTime("localparts", &middleTime);

    // groupReduceDistStreamEx read into partition vectors
      std::vector<struct239 > v242[nl1.partitionCount];
      
      dense_hash_map<int, int, hash<int >, eqint >::iterator v237;
      
    for (v237 = v236.begin(); v237 != v236.end(); ++v237) {
        int v240;
      struct239 v238;
        v238.v0 = (v237->first);
      v238.v1 = (v237->second);
        v240 = v238.v0 % nl1.partitionCount;
      
    v242[v240].push_back(v238);
    }
    // redistribute local hash maps
      std::vector<struct239 > v258;
      
    exchangePartitions<struct239 >(v242, &v258, &nl1);
    // read from recv buffs into local hash map
    v236.clear_no_resize();
      std::vector<struct239 >::iterator v259;
      
      int v232;
      
      int v233;
      
      int v234;
      
      int v235;
      
    for (v259=v258.begin(); v259!=v258.end(); v259++) {
        v232 = (*v259).v0;
      v233 = (*v259).v1;
      
      if (v236.count(v232)) {
        v234 = v236[v232];
          v235 = v233 + v234;
      
        v236[v232] = v235;
      } else {
        v236[v232] = v233;
      }
    }
     // finish readDistArr1D
  logTime("exchangedparts", &middleTime);

// readDistHashMap

/*if (NL_ON_FRINGE(nl1)) {
    dense_hash_map<int, int, hash<int >, eqint >::iterator v260;
    for (v260 = v236.begin(); v260 != v236.end(); ++v260) {
      printVal("(");
    printVal(v260->first);
    printVal(", ");
    printVal(v260->second);
    printVal(")");
    printVal("\n");
  }
}*/


nl1.Free();
}


int main(int argc, char **argv) {
  MPI::Init(argc, argv);
  MPI::Errhandler errHandler = MPI::COMM_WORLD.Create_errhandler(&throwErrHandler);
  MPI::COMM_WORLD.Set_errhandler(errHandler);  

  // seed random numbers
  srand (time(NULL));

  // init consts
  int Nh = 100;
  long N = 1073741824; // remember size is 8 times this = 10.5GB
  int Nproc = MPI::COMM_WORLD.Get_size();

  // the current date/time
  time_t timer = time(NULL);
  struct tm* dateTime = localtime(&timer);
  char *strTime = asctime(dateTime);
  char* strPtr = strTime;
  while (*strPtr != '\0') { if (*strPtr == '\n') *strPtr = '\0'; strPtr++; }
  
  // make log header
  snprintf(timeHeader, MAX_TIME_HEADER, "%s\t%li\t%i\t%i", strTime, N, Nh, Nproc);

  // time the computation
  initTime();
  run();
  logTime("total", &beginTime);

  // cleanup
  MPI::Finalize(); return 0;
}
